@@ -471,6 +471,12 @@ public class Leader {
                     f.ping();
                 }
 
+                // check leader running status
+                if (!this.isRunning()) {
+                    shutdown("Unexpected internal error");
+                    return;
+                }
+
               if (!tickSkip && !self.getQuorumVerifier().containsQuorum(syncedSet)) {
                 //if (!tickSkip && syncedCount < self.quorumPeers.size() / 2) {
                     // Lost quorum, shutdown
@@ -1055,4 +1061,8 @@ public class Leader {
             return "UNKNOWN";
         }
     }
+
+    private boolean isRunning() {
+        return self.isRunning() && zk.isRunning();
+    }
 }
